/* ２階正規型常微分方程式の解のグラフ    */
/*          difnorm2.c              */
/* (c) Kenrou Adachi, 2004/11/23,25 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <eggx.h>

#define BLUE   4
#define YELLOW 7


/* 関数宣言 */

void info_message();

void print(char *message);

void input_data(double *a, double *xo, double *yo, double *ts, double *te);

void view_window(double a, double ts, double te,
	       double *dx, double *dt, double *ox, double *ot);

void pre_draw(double xo, double ts, double dx, double dt,
	      double ox, double ot, int win);

void draw_axes(int win);

void draw_curve(double xo, double yo, double ts, double te,
		double dx, double dt, double ox, double ot, int win);

void runge_kutta(double *xo, double *yo, double *t, double h);

void draw_micro_line(double xo, double dx, double dt,
		     double ox, double ot, double t, int win);

/* 外部関数 */

extern double fx(double x, double y, double t);
extern double fy(double x, double y, double t);

/* 大域変数 */

float CPX, CPY;


/* トップルーチン */

int main(void)
{
  double a, xo, yo, ts, te;
  double dx, dt, ox, ot;
  int    win;

  /* 挨拶 */

  info_message();

  /* 変数の入力 */

  do
  {
    input_data(&a, &xo, &yo, &ts, &te);
  }
  while ((a < 0.0) || (a > 160.0) || (fabs(xo) >= fabs(a)) && (ts >= te));

  /* グラフィック初期化 */

  win = gopen(640, 400);
  gsetbgcolor(win, "black");
  gclr(win);

  /* 座標軸の描画 */

  draw_axes(win);

  /* 座標変換 */

  view_window(a, ts, te, &dx, &dt, &ox, &ot);

  /* 解の計算と描画 */

  draw_curve(xo, yo, ts, te, dx, dt, ox, ot, win);

  /* 終了処理 */

  fprintf(stderr, "終了しました。エンターキーを押して下さい。");
  getchar();
  getchar();
  gclose(win);

  return EXIT_SUCCESS;
}


/* utility */

void print(char *message)
{
  fprintf(stderr, message);
}


/* 挨拶 */

void info_message(void)
{
  print("このプログラムは２階正規型常微分方程式の解のグラフを描画するものです。\n");
  print("すなわち, d^2x/dt^2 = f(x, x', t) の解 x = x(t) のグラフです。\n");
}


/* データ入力 */

void input_data(double *a, double *xo, double *yo, double *ts, double *te)
{
  print(" x軸の描画範囲を入力して下さい。x : [-a, a]\n");
  print(" a = ? ");
  scanf("%lf", a);
  print("\n x の初期値 xo の値を入力して下さい。\n");
  print(" xo = ? ");
  scanf("%lf", xo);
  print("\n x' = dx/dt の初期値 dxo の値を入力して下さい。\n");
  print(" dxo = ? ");
  scanf("%lf", yo);
  print("\n 時間 t の変域を入力して下さい。t : [ts, te] です。\n");
  print(" ts = ? ");
  scanf("%lf", ts);
  print("\n te = ? ");
  scanf("%lf", te);
}


/* 座標変換前処理 */

void view_window(double a, double ts, double te,
		 double *dx, double *dt, double *ox, double *ot)
{
  *dx = 2.0 * a;
  *dt = te - ts;
  *ox = 400 * a / *dx;
  *ot = -640 * ts / *dt;
}


/* 座標軸の描画 */

void draw_axes(int win)
{
  newpen(win, BLUE);
  line(win,   0.0,   0.0, PENUP);
  line(win,   0.0, 399.0, PENDOWN);
  line(win,   0.0, 200.0, PENUP);
  line(win, 639.0, 200.0, PENDOWN);
}


/* 描画開始位置の設定 */

void pre_draw(double xo, double ts, double dx, double dt,
	      double ox, double ot, int win)
{
  double oldx, oldt;

  newpen(win, YELLOW);
  oldx =  -400.0 * xo / dx + ox;
  oldt =  640.0 * ts / dt + ot;
  line(win, (float)oldt, 400.0 - (float)oldx, PENUP);
  CPX = (float)oldt;
  CPY = (float)oldx;
}


/* Runge-Kutta 方で近似解を求めるルーチン */

void runge_kutta(double *xo, double *yo, double *t, double h)
{
  double x1, x2, x3, x4, y1, y2, y3, y4;

  x1 = h * fx(*xo, *yo, *t);
  y1 = h * fy(*xo, *yo, *t);
  x2 = h * fx(*xo + x1 / 2.0, *yo + y1 / 2.0, *t + h / 2.0);
  y2 = h * fy(*xo + x1 / 2.0, *yo + y1 / 2.0, *t + h / 2.0);
  x3 = h * fx(*xo + x2 / 2.0, *yo + y2 / 2.0, *t + h / 2.0);
  y3 = h * fy(*xo + x2 / 2.0, *yo + y2 / 2.0, *t + h / 2.0);
  x4 = h * fx(*xo + x3, *yo + y3, *t + h);
  y4 = h * fy(*xo + x3, *yo + y3, *t + h);

  *xo += (x1 + 2.0 * x2 + 2.0 * x3 + x4) / 6.0;
  *yo += (y1 + 2.0 * y2 + 2.0 * y3 + y4) / 6.0;
  *t  += h;
}


/* 線素の描画 */

void draw_micro_line(double xo, double dx, double dt,
		     double ox, double ot, double t, int win)
{
  line(win, CPX, 400.0 - CPY, PENUP);
  line(win, (float)(640.0 * t / dt + ot),
       400.0 - (float)(-400.0 * xo / dx + ox), PENDOWN);
  CPX = (float)(640.0 * t / dt + ot);
  CPY = (float)(-400.0 * xo / dx + ox);
}
 

/* 解曲線描画のメインルーチン */

void draw_curve(double xo, double yo, double ts, double te,
		double dx, double dt, double ox, double ot, int win)
{
  double t, h;
  int    i;

  h = (te - ts) / 640.0;
  t = ts;
  pre_draw(xo, t, dx, dt, ox, ot, win);

  for (i = 1; i <= 640; i++)
  {
    runge_kutta(&xo, &yo, &t, h);
    draw_micro_line(xo, dx, dt, ox, ot, t, win);
  }
}


/* 関数のサンプル */

/*
double fx(double x, double y, double t)
{
  return y;
}


double fy(double x, double y, double t)
{
  return (1.0 - x * x) * y - x;
}
*/
