/* ２階正規型常微分方程式の相空間のグラフ */
/*          difnorm2.c               */
/* (c) Kenrou Adachi, 2004/11/23     */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <eggx.h>

#define BLUE   4
#define YELLOW 7


void info_message();
void print(char *message);
void input_data(double *a, double *b, double *xo,
                double *yo, double *ts, double *te);
void view_window(double a, double b,
	       double *dx, double *dy, double *ox, double *oy);
void pre_draw(double xo, double yo, double dx, double dy,
	      double ox, double oy, int win);
void draw_axes(int win);
void draw_curve(double xo, double yo, double ts, double te,
		double dx, double dy, double ox, double oy, int win);
void runge_kutta(double *xo, double *yo, double *t, double h);
void draw_micro_line(double xo, double yo, double dx, double dy,
		     double ox, double oy, int win);

extern double fx(double x, double y, double t);
extern double fy(double x, double y, double t);

float CPX, CPY;

int main(void)
{
  double a, b, xo, yo, ts, te;
  double dx, dy, ox, oy;
  int    win;

  info_message();

  do
  {
    input_data(&a, &b, &xo, &yo, &ts, &te);
  }
  while ((a <= 0.0) || (a >= 256.0) || (fabs(xo) >= fabs(a))
	 || (ts >= te)  || (b <= 0.0) || (b >= 160.0));

  win = gopen(640, 400);
  gsetbgcolor(win, "black");
  gclr(win);

  draw_axes(win);
  view_window(a, b, &dx, &dy, &ox, &oy);
  draw_curve(xo, yo, ts, te, dx, dy, ox, oy, win);

  fprintf(stderr, "終了しました。エンターキーを押して下さい。");
  getchar();
  getchar();
  gclose(win);

  return EXIT_SUCCESS;
}


void print(char *message)
{
  fprintf(stderr, message);
}


void info_message(void)
{
  print("このプログラムは２階正規型常微分方程式の解の相空間における軌道を描画するものです。\n");
  print("すなわち, d^2x/dt^2 = f(x, x', t) の解 (x(t), x'(t))の軌道です。\n");
}


void input_data(double *a, double *b, double *xo,
                double *yo, double *ts, double *te)
{
  print(" x軸の描画範囲を入力して下さい。x : [-a, a]\n");
  print(" a = ? ");
  scanf("%lf", a);
  print(" y軸の描画範囲を入力して下さい。y : [-b, b]\n");
  print(" b = ? ");
  scanf("%lf", b);
  print("\n x の初期値 xo の値を入力して下さい。\n");
  print(" xo = ? ");
  scanf("%lf", xo);
  print("\n x' = dx/dt の初期値 yo の値を入力して下さい。\n");
  print(" yo = ? ");
  scanf("%lf", yo);
  print("\n 時間 t の変域を入力して下さい。t : [ts, te] です。\n");
  print(" ts = ? ");
  scanf("%lf", ts);
  print("\n te = ? ");
  scanf("%lf", te);
}


void view_window(double a, double b,
		 double *dx, double *dy, double *ox, double *oy)
{
  *dx = 2.0 * a;
  *dy = 2.0 * b;
  *ox = 640 * a / *dx;
  *oy = 400 * b / *dy;
}


void draw_axes(int win)
{
  newpen(win, BLUE);
  line(win, 320.0,   0.0, PENUP);
  line(win, 320.0, 399.0, PENDOWN);
  line(win,   0.0, 200.0, PENUP);
  line(win, 639.0, 200.0, PENDOWN);
}


void pre_draw(double xo, double yo, double dx, double dy,
	      double ox, double oy, int win)
{
  double oldx, oldy;

  newpen(win, YELLOW);
  oldx =  640.0 * xo / dx + ox;
  oldy =  -400.0 * yo / dy + oy;
  line(win, (float)oldx, 400.0 - (float)oldy, PENUP);
  CPX = (float)oldx;
  CPY = (float)oldy;
}


void runge_kutta(double *xo, double *yo, double *t, double h)
{
  double x1, x2, x3, x4, y1, y2, y3, y4;

  x1 = h * fx(*xo, *yo, *t);
  y1 = h * fy(*xo, *yo, *t);
  x2 = h * fx(*xo + x1 / 2.0, *yo + y1 / 2.0, *t + h / 2.0);
  y2 = h * fy(*xo + x1 / 2.0, *yo + y1 / 2.0, *t + h / 2.0);
  x3 = h * fx(*xo + x2 / 2.0, *yo + y2 / 2.0, *t + h / 2.0);
  y3 = h * fy(*xo + x2 / 2.0, *yo + y2 / 2.0, *t + h / 2.0);
  x4 = h * fx(*xo + x3, *yo + y3, *t + h);
  y4 = h * fy(*xo + x3, *yo + y3, *t + h);

  *xo += (x1 + 2.0 * x2 + 2.0 * x3 + x4) / 6.0;
  *yo += (y1 + 2.0 * y2 + 2.0 * y3 + y4) / 6.0;
  *t  += h;
}


void draw_micro_line(double xo, double yo, double dx, double dy,
		     double ox, double oy, int win)
{
  line(win, CPX, 400.0 - CPY, PENUP);
  line(win, (float)(640.0 * xo / dx + ox),
       400.0 - (float)(-400.0 * yo / dy + oy), PENDOWN);
  CPX = (float)(640.0 * xo / dx + ox);
  CPY = (float)(-400.0 * yo / dy + oy);
}
 

void draw_curve(double xo, double yo, double ts, double te,
		double dx, double dy, double ox, double oy, int win)
{
  double t, h;
  int    i;

  h = (te - ts) / 640.0;
  t = ts;
  pre_draw(xo, yo, dx, dy, ox, oy, win);

  for (i = 1; i <= 640; i++)
  {
    runge_kutta(&xo, &yo, &t, h);
    draw_micro_line(xo, yo, dx, dy, ox, oy, win);
  }
}

/*

double fx(double x, double y, double t)
{
  return y;
}


double fy(double x, double y, double t)
{
  return (1.0 - x * x) * y - x;
}

*/
